//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

use crate::simu;

pub struct Gravity {}

impl Gravity {
    pub fn new() -> Self {
        Self {}
    }
}

impl simu::Animator for Gravity {
    fn animate(
        &self,
        entity: &mut simu::Entity,
        _window: &simu::Window,
        dt: f64,
    ) {
        let kinetics = entity.kinetics_mut();
        kinetics.y_speed += 10.0 * 9.81 * dt;
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Friction {
    coef: f64,
}

impl Friction {
    pub fn new(coef: f64) -> Self {
        Self { coef }
    }
}

impl simu::Animator for Friction {
    fn animate(
        &self,
        entity: &mut simu::Entity,
        _window: &simu::Window,
        dt: f64,
    ) {
        let kinetics = entity.kinetics_mut();
        kinetics.friction(self.coef, dt);
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Bounce {
    elasticity: f64,
}

impl Bounce {
    pub fn new(elasticity: f64) -> Self {
        Self { elasticity }
    }
}

impl simu::Animator for Bounce {
    fn animate(
        &self,
        entity: &mut simu::Entity,
        window: &simu::Window,
        _dt: f64,
    ) {
        let window_size = window.size();
        let radius = entity.bounding_radius();
        let kinetics = entity.kinetics_mut();
        if kinetics.x < radius {
            kinetics.bounce(1.0, 0.0, self.elasticity);
        } else if kinetics.x > window_size.w - radius {
            kinetics.bounce(-1.0, 0.0, self.elasticity);
        }
        if kinetics.y < radius {
            kinetics.bounce(0.0, 1.0, self.elasticity);
        } else if kinetics.y > window_size.h - radius {
            kinetics.bounce(0.0, -1.0, self.elasticity);
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Stop {}

impl Stop {
    pub fn new() -> Self {
        Self {}
    }
}

impl simu::Animator for Stop {
    fn animate(
        &self,
        entity: &mut simu::Entity,
        window: &simu::Window,
        _dt: f64,
    ) {
        let window_size = window.size();
        let radius = entity.bounding_radius();
        let max_x = window_size.w - radius;
        let max_y = window_size.h - radius;
        let kinetics = entity.kinetics_mut();
        kinetics.x = kinetics.x.clamp(radius, max_x);
        kinetics.y = kinetics.y.clamp(radius, max_y);
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Pair {
    first: Box<dyn simu::Animator>,
    second: Box<dyn simu::Animator>,
}

impl Pair {
    pub fn new(
        first: Box<dyn simu::Animator>,
        second: Box<dyn simu::Animator>,
    ) -> Self {
        Self { first, second }
    }
}

impl simu::Animator for Pair {
    fn animate(
        &self,
        entity: &mut simu::Entity,
        window: &simu::Window,
        dt: f64,
    ) {
        self.first.animate(entity, window, dt);
        self.second.animate(entity, window, dt);
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Repulse {
    coef: f64,
}

impl Repulse {
    pub fn new(coef: f64) -> Self {
        Self { coef }
    }
}

impl simu::Animator for Repulse {
    fn animate(
        &self,
        entity: &mut simu::Entity,
        window: &simu::Window,
        dt: f64,
    ) {
        let radius = entity.bounding_radius();
        for other_cell in window.entities().iter() {
            if !std::ptr::eq(entity, other_cell.as_ptr()) {
                let mut other = other_cell.borrow_mut();
                let other_radius = other.bounding_radius();
                let other_kinetics = other.kinetics_mut();
                let kinetics = entity.kinetics_mut();
                kinetics.repulse(
                    other_kinetics,
                    radius + other_radius,
                    self.coef,
                    dt,
                );
            }
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Collide {
    elasticity: f64,
}

impl Collide {
    pub fn new(elasticity: f64) -> Self {
        Self { elasticity }
    }
}

impl simu::Animator for Collide {
    fn animate(
        &self,
        entity: &mut simu::Entity,
        window: &simu::Window,
        _dt: f64,
    ) {
        let radius = entity.bounding_radius();
        for other_cell in window.entities().iter() {
            if !std::ptr::eq(entity, other_cell.as_ptr()) {
                let mut other = other_cell.borrow_mut();
                let other_radius = other.bounding_radius();
                let other_kinetics = other.kinetics_mut();
                let kinetics = entity.kinetics_mut();
                if kinetics.distance(other_kinetics) < radius + other_radius {
                    kinetics.collide(
                        other_kinetics,
                        radius * radius * radius,
                        other_radius * other_radius * other_radius,
                        self.elasticity,
                    );
                }
            }
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
