//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#![deny(warnings)]
#![allow(dead_code)]
#![allow(clippy::new_without_default)]

pub mod animators;
pub mod interactors;
pub mod s5;
pub mod shapes;
pub mod simu;

pub type ApplicationWindow = simu::Window;

fn create_window(
    args: &[&str]
) -> Result<ApplicationWindow, Box<dyn std::error::Error>> {
    println!("args: {:?}", args);
    let mut window = simu::Window::new(s5::Rect { w: 800.0, h: 600.0 });
    //~~~~ populate window ~~~~
    use std::f64::consts::PI;
    let mut rng = s5::Rng::new();
    for i in 0..16 {
        //~~~~ determine initial state ~~~~
        let kinetics = s5::KineticState {
            x: rng.range(0.0, window.size().w),
            y: rng.range(0.0, window.size().h),
            angle: rng.range(0.0, 2.0 * PI),
            x_speed: rng.sign::<f64>() * rng.range(50.0, 100.0),
            y_speed: rng.sign::<f64>() * rng.range(50.0, 100.0),
            angular_speed: rng.sign::<f64>() * rng.range(PI / 6.0, PI / 2.0),
        };
        let color = s5::Color::new(rng.number(), rng.number(), rng.number());

        //~~~~ choose shape ~~~~
        let filled = rng.number();
        let shape: Box<dyn simu::Shape> = match i % 2 == 0 {
            true => {
                Box::new(shapes::Circle::new(rng.range(25.0, 50.0), filled))
            }
            false => Box::new(shapes::Rectangle::new(
                s5::Rect {
                    w: rng.range(50.0, 100.0),
                    h: rng.range(50.0, 100.0),
                },
                filled,
            )),
        };

        //~~~~ choose animator ~~~~
        let animator: Box<dyn simu::Animator> = match filled {
            true => Box::new(animators::Pair::new(
                Box::new(animators::Gravity::new()),
                Box::new(animators::Pair::new(
                    Box::new(animators::Bounce::new(0.75)),
                    Box::new(animators::Collide::new(0.75)),
                )),
            )),
            // false => Box::new(animators::Stop::new()),
            false => Box::new(animators::Pair::new(
                Box::new(animators::Repulse::new(100.0)),
                Box::new(animators::Friction::new(0.1)),
            )),
        };

        //~~~~ choose interactor ~~~~
        let interactor: Box<dyn simu::Interactor> = match filled {
            true => Box::new(interactors::Pair::new(
                Box::new(interactors::Push::new(500.0)),
                Box::new(interactors::Chromo::new()),
            )),
            false => Box::new(interactors::Pair::new(
                Box::new(interactors::FastForward::new(0.5)),
                Box::new(interactors::Turn::new()),
            )),
        };

        //~~~~ create entity ~~~~
        window.add_entity(simu::Entity::new(
            kinetics,
            color,
            shape,
            Some(animator),
            Some(interactor),
        ));
    }
    Ok(window)
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
